# microsoft_com.py

# Copyright (C) 2003 Phillip Pearson
#   http://www.myelin.co.nz/microsoft_com_py/

# Released under the same license as Python.  Share and enjoy.

# This code is _not_ derived from the sample code released in the
# Microsoft SDK, so it seems to me that it should be fine to
# distribute in a cross platform form.  I am not a lawyer; if you are,
# feel free to contact me and give me advice :-)

# Bug reports and enhancements are welcome; you can contact me by
# going to the project homepage and clicking on the 'contact' link.

# USAGE:

# First, you need a developer's token.  Go to
# http://msdn.microsoft.com/webservices/building/livewebservices/mscomservices/default.aspx
# and get one.  Now create a file called mstoken.py and put your token
# and password in it like so:

#   token = '2346iusdflkjah9873562'
#   pin = 'aksjdhfliaugw34'

# Now you can import microsoft_com and do things.  Look at the
# docstrings for GetVersion(), GetCultures(), GetTopDownloads() and
# GetDownloadDetail() for more information.  Also look at the bottom
# of this file for some example usage.

# The SOAP endpoint also has some documentation if you open it in a
# web browser: http://ws.microsoft.com/MsComService/MsCom.asmx

__LIBRARY__ = ('microsoft_com.py', 'http://www.myelin.co.nz/microsoft_com_py/')
__AUTHOR__ = ('Phillip Pearson', 'http://www.myelin.co.nz/post/')
__LICENSE__ = 'Python'

# system bits
import binascii
import httplib
import random
import sha
from StringIO import StringIO
import sys
import time

# for the SOAP headers
from elementtree.ElementTree import Element, SubElement, ElementTree

# for the SOAP itself
from ZSI import TC
from ZSI.client import Binding

# your user token and PIN
import mstoken

# set this to 1 for a LOT of debugging information
verbose = 0

def _dump(xml, indent=''):
    "dump a DOM tree"
    print indent + `xml`
    for child in xml.childNodes:
        _dump(child, indent=indent + '    ')

def _call(opname, params):
    "call a microsoft.com function"
    if verbose:
        print "----------------------------"
        print "calling %s" % opname
        print "----------------------------"
    NS = "http://www.microsoft.com"
    server = Binding(host="ws.microsoft.com",
                     port=80,
                     soapaction=NS + '/' + opname,
                     ns=NS,
                     url="/MsComService/MsCom.asmx",
                     )
    if verbose:
        server.trace = sys.stdout

    wsse = Element('wsse:Security')
    wsse.set('xmlns:wsse', 'http://schemas.xmlsoap.org/ws/2002/07/secext')
    wsse.set('xmlns:wsu', 'http://schemas.xmlsoap.org/ws/2002/07/utility')
    ut = SubElement(wsse, 'wsse:UsernameToken')
    def mkelem(parent, tagname, text, **attrs):
        if tagname.find(':') == -1:
            tagname = 'wsse:' + tagname
        x = SubElement(parent, tagname)
        x.text = text
        for k,v in attrs.items():
            x.set(k, v)
    mkelem(ut, 'Username', mstoken.token)
    created = time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime())
    nonce = sha.new(str(random.random())).digest()
    digest = sha.new(nonce + created + mstoken.pin).digest()
    mkelem(ut, 'Password', binascii.b2a_base64(digest), Type='wsse:PasswordDigest')
    mkelem(ut, 'Nonce', binascii.b2a_base64(nonce))
    mkelem(ut, 'wsu:Created', created)

    tree = ElementTree(wsse)
    wsse = StringIO()
    tree.write(wsse)
    wsse_header = wsse.getvalue()

    rpc_ret = server.RPC(None, opname, params, auth_header=wsse_header)
    if verbose:
        print rpc_ret
        print server.ps
        _dump(server.ps.body)

    return server.ps


def _get_node_text(node):
    "pull out all the text under a DOM node"
    text = ''
    for c in node.childNodes:
        if c.nodeType == c.TEXT_NODE:
            text += c.nodeValue
    return text

class SoapClass:
    "base for all the structs that are decoded from SOAP responses"
    def __init__(self, node):
        if verbose: print "%s.__init__" % self.__class__.__name__
        for x in self.keys:
            if x.startswith('+'): x = x[1:]
            setattr(self, x, None)
        got = 0
        for child in node.childNodes:
            if child.nodeName in self.keys:
                setattr(self, child.nodeName, _get_node_text(child))
            elif '+' + child.nodeName in self.keys:
                if verbose: print "found + key (%s); copying node (%s) in there instead" % (child.nodeName, `child`)
                setattr(self, child.nodeName, child)
            else:
                if verbose: print "unknown tag in xml for %s: %s" % (self.__class__.__name__, child.nodeName)
    def __str__(self):
        values = []
        for k in self.keys:
            if k.startswith('+'):
                k = k[1:]
            if hasattr(self, k):
                values.append('%s=%s' % (k, `getattr(self, k)`))
            else:
                values.append('%s!' % k)
        return "<%s at %x: %s>" % (
            self.__class__.__name__,
            id(self),
            ", ".join(values))
    def __repr__(self):
        return self.__str__()

class Culture(SoapClass):
    keys = ('CultureID', 'Name', 'Description')

class DownloadSummary(SoapClass):
    keys = ("DownloadID",
            "CultureID",
            "ShortName",
            "ShortDescription",
            "DatePublished",
            "RankOverall",
            "RankInThisCulture",
            "DetailsUrl",)

def _zoom(node, levels):
    "walk down a path of given elements in a DOM tree"
    for level in levels:
        got_one = 0
        for child in node.childNodes:
            if child.nodeName == level:
                node = child
                got_one = 1
                break
        if not got_one:
            raise Exception("expected <%s> attribute; couldn't find it!" % level)
    return node

def GetVersion():
    """

    Retrieves version information for the web service.

    """
    ret = _call('GetVersion', ())
    version = _zoom(ret.body, ('GetVersionResponse', 'GetVersionResult'))
    return _get_node_text(version)

def GetCultures():
    """

    Retrieves a list of cultures and identifiers that can be passed
    into other methods.

    """
    cultures = _call('GetCultures', ())
    result = _zoom(cultures.body, ('GetCulturesResponse', 'GetCulturesResult'))
    ret = []
    for culture in result.childNodes:
        ret.append(Culture(culture))
    return ret

class GetTopDownloadsParams:
    pass
GetTopDownloadsParams.typecode = TC.Struct(GetTopDownloadsParams, (
    TC.String('topType'),
    TC.Integer('topN'),
    TC.String('cultureID'),
    ), 'GetTopDownloads')

def GetTopDownloads(toptype, topn, cultureid):
    """

    Retrieves information on the top downloads under certain criteria.

    toptype can be 'Popular' or 'Recent'

    topn is a number from 1 to 25

    cultureid is the ID of the culture for which you want to receive download info

    """
    params = GetTopDownloadsParams()
    params.topN = topn
    params.topType = toptype
    params.cultureID = cultureid
    topdls = _call('GetTopDownloads', params)
    ret = []
    for d in _zoom(topdls.body, ('GetTopDownloadsResponse', 'GetTopDownloadsResult')).childNodes:
        ret.append(DownloadSummary(d))
    return ret

class GetDownloadDetailParams:
    pass
GetDownloadDetailParams.typecode = TC.Struct(GetDownloadDetailParams, (
    TC.String('downloadID'),
    TC.String('cultureID'),
    ), 'GetDownloadDetail')

class DownloadDetail(SoapClass):
    keys = ('DownloadID',
            'CultureID',
            'DatePublished',
            'Version',
            'Name',
            'ShortName',
            'SubTitle',
            'Keywords',
            'MinTotalSize',
            'MaxTotalSize',
            'Requirements',
            'Instructions',
            'Description',
            'ShortDescription',
            'AdditionalInfo',
            '+RelatedLinks',
            '+DCCategories',
            '+RelatedProductsOrTechnologies',
            '+Releases',
            '+RelatedDownloads',
            )

class RelatedLink(SoapClass):
    keys = ('Description',
            'Url',
            )

class DCCategory(SoapClass):
    keys = ('Name',
            )

class ProductOrTechnology(SoapClass):
    keys = ('ShortName',
            'LegalName',
            )

class Release(SoapClass):
    keys = ('ID',
            '+SupportedLanguages',
            '+SupportedOSs',
            '+Files',
            )

class SupportedLanguage(SoapClass):
    keys = ('Name',
            )

class SupportedOS(SoapClass):
    keys = ('Name',
            )

class File(SoapClass):
    keys = ('Name',
            'Url',
            'Size',
            'Display',
            'IsCombined',
            )

class RelatedDownload(SoapClass):
    keys = ('DownloadID',
            'ShortName',
            )
    
def _populate_groups(soapobj, groups):
    "being too lazy to define the responses all in one place, we define one layer of each struct at a time, and call _populate_groups lots of times when we have to decode nested structs"
    if verbose: print "--- populate_groups (obj=%s)" % `soapobj`
    for group,clas in groups:
        if verbose: print "    populate_groups; group=%s, class=%s" % (group,clas)
        node = getattr(soapobj, group)
        if verbose: print node
        list = []
        if node is not None:
            for child in node.childNodes:
                if verbose: print "        turning child (%s) into a %s!" % (`child`, clas.__name__)
                list.append(clas(child))
        setattr(soapobj, group, list)

def GetDownloadDetail(downloadid, cultureid):
    """

    Retrieves detailed information about a particular download for a
    particular culture.

    downloadid is the ID of the download (from GetTopDownloads())

    cultureid is the ID of the culture you are looking at (from
    GetCultures() -- you should have passed it in to GetTopDownloads()
    as well)
    
    """
    params = GetDownloadDetailParams
    params.downloadID = downloadid
    params.cultureID = cultureid
    detail = _call('GetDownloadDetail', params)
    ret = []
    dd = DownloadDetail(_zoom(detail.body, ('GetDownloadDetailResponse', 'GetDownloadDetailResult')))
    _populate_groups(dd, (('RelatedLinks', RelatedLink),
                         ('DCCategories', DCCategory),
                         ('RelatedProductsOrTechnologies', ProductOrTechnology),
                         ('Releases', Release),
                         ('RelatedDownloads', RelatedDownload),
                         ))
    for rel in dd.Releases:
        _populate_groups(rel, (('SupportedLanguages', SupportedLanguage),
                              ('SupportedOSs', SupportedOS),
                              ('Files', File),
                              ))
            
    return dd

# testing stuff
if __name__ == '__main__':
    print "get version ..."
    print GetVersion()
    print "get cultures ..."
    cultures = GetCultures()
    for culture in cultures:
        print "culture:",culture
    cultureid=cultures[0].CultureID
    print "get top downloads for",cultureid
    tops = GetTopDownloads('Popular', 20, cultureid)
    print tops
    for top in tops:
        print "get details for",top.DownloadID
        print GetDownloadDetail(top.DownloadID, cultureid)
